import cv2
import numpy as np

from face.src.face_detection import RetinaFace
from face.src.face_embedding import ArcFace
from face.src.landmark import Landmark
from tracking.src.byte_tracker import BYTETracker

retina = RetinaFace()
arc = ArcFace()
lan = Landmark()
bt = BYTETracker()

img = cv2.imread('/media/vti/SSD/VTI/FaceMask/face_recognize_v2/5.png')
boxes, kpss = retina.detect(img)

for b,k in zip(boxes, kpss):
    emb = arc.get(img, k)
    print(emb)
    l = lan.get(img, b)
    print(l)


while True:
    frame = img
    boxes,_ = retina.detect(frame)
    boxes, ids = bt.predict(frame, boxes)







from time import time
from click import argument
import cv2
from cv2 import waitKey
import torch, os, time
import numpy as np
import onnxruntime as ort

from .models.common import DetectMultiBackend
from .utils.augmentations import letterbox
from .utils.general import (LOGGER, check_img_size,scale_coords, cv2, non_max_suppression)


class Yolov5():
    def __init__(self, pt_model_path, yaml_path, imgsz=(640,640), augment=False,conf_thres=0.5, iou_thres=0.45, classes=None, agnostic_nms=False, max_det=100) -> None:
        self.imgsz = imgsz
        self.augment = augment
        self.conf_thres = conf_thres
        self.iou_thres = iou_thres
        self.classes = classes
        self.agnostic_nms = agnostic_nms
        self.max_det = max_det
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.model = DetectMultiBackend(pt_model_path, device=self.device, dnn=False, data=yaml_path, fp16=False)
        self.stride, self.names, self.pt = self.model.stride, self.model.names, self.model.pt
        self.imgsz = check_img_size(imgsz, s=self.stride)  # check image size

    def inference(self, img):
        h,w = img.shape[0:2]
        gain = max(w/self.imgsz[0], h/self.imgsz[1])
        img = letterbox(img, self.imgsz, stride=self.stride, auto=self.pt)[0]
        im = img.transpose((2, 0, 1))[::-1]  # HWC to CHW, BGR to RGB
        im = np.ascontiguousarray(im)
        im = torch.from_numpy(im).to(self.device)
        im = im.half() if self.model.fp16 else im.float()
        im /= 255  # 0 - 255 to 0.0 - 1.0
        if len(im.shape) == 3:
            im = im[None] 
        # visualize = increment_path(save_dir / Path(path).stem, mkdir=True) if visualize else False
        pred = self.model(im, augment=self.augment)
        pred = non_max_suppression(pred, self.conf_thres, self.iou_thres, self.classes, self.agnostic_nms, max_det=self.max_det)

        pred = pred[0].cpu().detach().numpy()
        for i, pre in enumerate(pred):
            pred[i][:4] = np.array(pre[:4]*gain)

        # pred[:, :4] = scale_coords(im.shape[2:], pred[:, :4], img.shape).round()

        return pred


class Yolov5():
    def __init__(self, model_path, imw=640, imh=640, conf_thres=0.5, iou_thres=0.45, max_det=100) -> None:
        self.imw = imw
        self.imh = imh
        self.conf_thres=conf_thres
        self.iou_thres=iou_thres
        self.max_det=max_det
        self.model = ort.InferenceSession(model_path)
        self.input_name = self.model.get_inputs()[0].name

    def inference(self, image):
        h,w = image.shape[0:2]
        rw = w/self.imw
        rh = h/self.imh
        img = cv2.resize(image,(self.imw,self.imh))
        img = img.transpose((2, 0, 1))[::-1]
        img = img.astype('float32') / 255.0
        img = np.expand_dims(img, axis=0)

        pred = self.model.run(None, {self.input_name: img.astype(np.float32)})
        pred = torch.FloatTensor(pred)

        pred = non_max_suppression(pred, conf_thres=self.conf_thres, iou_thres=self.iou_thres, classes=None, agnostic=False, max_det=self.max_det)
        pred = pred[0].cpu().detach().numpy()

        for i, pre in enumerate(pred):
            pred[i][:4] = np.array(pre[:4]*[rw,rh,rw,rh])

        return pred



from configparser import ConfigParser
import cv2, math, os
from PIL import Image, ImageDraw
from cv2 import mean
from matplotlib.pyplot import axis
from tqdm import tqdm
import random
from imutils import face_utils
import requests
from zipfile import ZipFile
from tqdm import tqdm
import bz2, shutil
import numpy as np



def get_line(face_landmark, image, type="eye", debug=False):
    pil_image = Image.fromarray(image)
    d = ImageDraw.Draw(pil_image)
    left_eye = face_landmark["left_eye"]
    right_eye = face_landmark["right_eye"]
    left_eye_mid = np.mean(np.array(left_eye), axis=0)
    right_eye_mid = np.mean(np.array(right_eye), axis=0)
    eye_line_mid = (left_eye_mid + right_eye_mid) / 2

    if type == "eye":
        left_point = left_eye_mid
        right_point = right_eye_mid
        mid_point = eye_line_mid

    elif type == "nose_mid":
        nose_length = (
            face_landmark["nose_bridge"][-1][1] - face_landmark["nose_bridge"][0][1]
        )
        left_point = [left_eye_mid[0], left_eye_mid[1] + nose_length / 2]
        right_point = [right_eye_mid[0], right_eye_mid[1] + nose_length / 2]
        # mid_point = (
        #     face_landmark["nose_bridge"][-1][1] + face_landmark["nose_bridge"][0][1]
        # ) / 2

        mid_pointY = (
            face_landmark["nose_bridge"][-1][1] + face_landmark["nose_bridge"][0][1]
        ) / 2
        mid_pointX = (
            face_landmark["nose_bridge"][-1][0] + face_landmark["nose_bridge"][0][0]
        ) / 2
        mid_point = (mid_pointX, mid_pointY)

    elif type == "nose_tip":
        nose_length = (
            face_landmark["nose_bridge"][-1][1] - face_landmark["nose_bridge"][0][1]
        )
        left_point = [left_eye_mid[0], left_eye_mid[1] + nose_length]
        right_point = [right_eye_mid[0], right_eye_mid[1] + nose_length]
        mid_point = (
            face_landmark["nose_bridge"][-1][1] + face_landmark["nose_bridge"][0][1]
        ) / 2

    elif type == "bottom_lip":
        bottom_lip = face_landmark["bottom_lip"]
        bottom_lip_mid = np.max(np.array(bottom_lip), axis=0)
        shiftY = bottom_lip_mid[1] - eye_line_mid[1]
        left_point = [left_eye_mid[0], left_eye_mid[1] + shiftY]
        right_point = [right_eye_mid[0], right_eye_mid[1] + shiftY]
        mid_point = bottom_lip_mid

    elif type == "perp_line":
        bottom_lip = face_landmark["bottom_lip"]
        bottom_lip_mid = np.mean(np.array(bottom_lip), axis=0)

        left_point = eye_line_mid
        left_point = face_landmark["nose_bridge"][0]
        right_point = bottom_lip_mid

        mid_point = bottom_lip_mid

    elif type == "nose_long":
        nose_bridge = face_landmark["nose_bridge"]
        left_point = [nose_bridge[0][0], nose_bridge[0][1]]
        right_point = [nose_bridge[-1][0], nose_bridge[-1][1]]

        mid_point = left_point

    # d.line(eye_mid, width=5, fill='red')
    y = [left_point[1], right_point[1]]
    x = [left_point[0], right_point[0]]
    # cv2.imshow('h', image)
    # cv2.waitKey(0)
    eye_line = fit_line(x, y, image)
    d.line(eye_line, width=5, fill="blue")

    # Perpendicular Line
    # (midX, midY) and (midX - y2 + y1, midY + x2 - x1)
    y = [
        (left_point[1] + right_point[1]) / 2,
        (left_point[1] + right_point[1]) / 2 + right_point[0] - left_point[0],
    ]
    x = [
        (left_point[0] + right_point[0]) / 2,
        (left_point[0] + right_point[0]) / 2 - right_point[1] + left_point[1],
    ]
    perp_line = fit_line(x, y, image)
    if debug:
        d.line(perp_line, width=5, fill="red")
        pil_image.show()
    return eye_line, perp_line, left_point, right_point, mid_point


def get_points_on_chin(line, face_landmark, chin_type="chin"):
    chin = face_landmark[chin_type]
    points_on_chin = []
    for i in range(len(chin) - 1):
        chin_first_point = [chin[i][0], chin[i][1]]
        chin_second_point = [chin[i + 1][0], chin[i + 1][1]]

        flag, x, y = line_intersection(line, (chin_first_point, chin_second_point))
        if flag:
            points_on_chin.append((x, y))

    return points_on_chin


def plot_lines(face_line, image, debug=False):
    pil_image = Image.fromarray(image)
    if debug:
        d = ImageDraw.Draw(pil_image)
        d.line(face_line, width=4, fill="white")
        pil_image.show()


def line_intersection(line1, line2):
    # mid = int(len(line1) / 2)
    start = 0
    end = -1
    line1 = ([line1[start][0], line1[start][1]], [line1[end][0], line1[end][1]])

    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])
    x = []
    y = []
    flag = False

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        return flag, x, y

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div

    segment_minX = min(line2[0][0], line2[1][0])
    segment_maxX = max(line2[0][0], line2[1][0])

    segment_minY = min(line2[0][1], line2[1][1])
    segment_maxY = max(line2[0][1], line2[1][1])

    if (
        segment_maxX + 1 >= x >= segment_minX - 1
        and segment_maxY + 1 >= y >= segment_minY - 1
    ):
        flag = True

    return flag, x, y


def fit_line(x, y, image):
    if x[0] == x[1]:
        x[0] += 0.1
    coefficients = np.polyfit(x, y, 1)
    polynomial = np.poly1d(coefficients)
    x_axis = np.linspace(0, image.shape[1], 50)
    y_axis = polynomial(x_axis)
    eye_line = []
    for i in range(len(x_axis)):
        eye_line.append((x_axis[i], y_axis[i]))

    return eye_line


def get_angle(line1, line2):
    delta_y = line1[-1][1] - line1[0][1]
    delta_x = line1[-1][0] - line1[0][0]
    perp_angle = math.degrees(math.atan2(delta_y, delta_x))
    if delta_x < 0:
        perp_angle = perp_angle + 180
    if perp_angle < 0:
        perp_angle += 360
    if perp_angle > 180:
        perp_angle -= 180

    # print("perp", perp_angle)
    delta_y = line2[-1][1] - line2[0][1]
    delta_x = line2[-1][0] - line2[0][0]
    nose_angle = math.degrees(math.atan2(delta_y, delta_x))

    if delta_x < 0:
        nose_angle = nose_angle + 180
    if nose_angle < 0:
        nose_angle += 360
    if nose_angle > 180:
        nose_angle -= 180
    # print("nose", nose_angle)

    angle = nose_angle - perp_angle
    return angle




def draw_landmarks(face_landmarks, image):
    pil_image = Image.fromarray(image)
    d = ImageDraw.Draw(pil_image)
    for facial_feature in face_landmarks.keys():
        d.line(face_landmarks[facial_feature], width=5, fill="white")
    pil_image.show()



def get_avg_brightness(img):
    img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(img_hsv)
    return np.mean(v)


def get_avg_saturation(img):
    img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(img_hsv)
    return np.mean(v)


def change_brightness(img, value=1.0):
    img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(img_hsv)
    v = value * v
    v[v > 255] = 255
    v = np.asarray(v, dtype=np.uint8)
    final_hsv = cv2.merge((h, s, v))
    img = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR)
    return img


def change_saturation(img, value=1.0):
    img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(img_hsv)
    s = value * s
    s[s > 255] = 255
    s = np.asarray(s, dtype=np.uint8)
    final_hsv = cv2.merge((h, s, v))
    img = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR)
    return img


def check_path(path):
    is_directory = False
    is_file = False
    is_other = False
    if os.path.isdir(path):
        is_directory = True
    elif os.path.isfile(path):
        is_file = True
    else:
        is_other = True

    return is_directory, is_file, is_other



def convert_106p_to_86p(pt106):
    landmark106to68=[1,10,12,14,16,3,5,7,0,23,21,19,32,30,28,26,17,    # 脸颊17点
                 43,48,49,51,50,      # 左眉毛5点
                 102,103,104,105,101, # 右眉毛5点
                 72,73,74,86,78,79,80,85,84, # 鼻子9点
                 35,41,42,39,37,36,   # 左眼睛6点
                 89,95,96,93,91,90,   # 右眼睛6点
                 52,64,63,71,67,68,61,65,66,62,70,69, # 嘴巴20点
                 61,58,59,53,56,55,52,69,57,60,54,65 #bottom_lip
                 ]
    pt86 = {'left_eyebrow':[], 'right_eyebrow':[], 'nose_bridge':[], 'nose_tip':[], 'left_eye':[], 'right_eye':[], 'top_lip':[], 'bottom_lip':[], 'chin':[]}
    for i in range(68):
        if i < 17:
            index = landmark106to68[i]
            pt86['chin'].append(pt106[index])
        elif i<22:
            index = landmark106to68[i]
            pt86['left_eyebrow'].append(pt106[index])
        elif i<27:
            index = landmark106to68[i]
            pt86['right_eyebrow'].append(pt106[index])
        elif i<31:
            index = landmark106to68[i]
            pt86['nose_bridge'].append(pt106[index])
        elif i<36:
            index = landmark106to68[i]
            pt86['nose_tip'].append(pt106[index])
        elif i<42:
            index = landmark106to68[i]
            pt86['left_eye'].append(pt106[index])
        elif i<48:
            index = landmark106to68[i]
            pt86['right_eye'].append(pt106[index])
        elif i<60:
            index = landmark106to68[i]
            pt86['top_lip'].append(pt106[index])
        else:
            index = landmark106to68[i]
            pt86['bottom_lip'].append(pt106[index])

    return pt86


def is_image(path):
    try:
        extensions = path[-4:]
        image_extensions = ["png", "PNG", "jpg", "JPG"]

        if extensions[1:] in image_extensions:
            return True 
        else:
            print("Please input image file. png / jpg")
            return False 
    except: 
        return False 





import cv2
import numpy as np

from . import face_align
import onnx
import onnxruntime
# import tensorflow as tf
import cv2
from sklearn.preprocessing import normalize
import os, gdown

model_dir = os.path.join(os.path.expanduser('~'), '.hawkeye/model')
class ArcFace:
    def __init__(self, model_name='', model_file=None, session=None):

        '''
        model_name: choice in ['arcface_s', 'arcface_m', 'arcface_l']
                    default is arcface_m
                    arcface_s -> SCRFD-500MF
                    arcface_m -> SCRFD-2.5GF
                    arcface_l -> SCRFD-10GF
        '''
        if model_file is None:
            assert model_name in ['', 'arcface_s', 'arcface_m', 'arcface_l'], 'model_name is wrong'
            if model_name == 'arcface_s':
                self.model_file = model_dir + '/s_w600k_mbf.onnx'
                if os.path.exists(model_dir + '/s_w600k_mbf.onnx') == False:
                    gdown.download('https://drive.google.com/u/0/uc?id=1Dvxphqyc84urh001xFPUWxmhVvrI7IbS&export=download', model_dir + '/s_w600k_mbf.onnx', quiet=False)
            if model_name == 'arcface_m' or  model_name == '':
                self.model_file = model_dir + '/glint360_r50.onnx'
                if os.path.exists(model_dir + '/glint360_r50.onnx') == False:
                    gdown.download('https://drive.google.com/u/0/uc?id=1obFVt88uEuoOYkOV2G6oAoQdBwSJGgy2&export=download', model_dir + '/glint360_r50.onnx', quiet=False)
            if model_name == 'arcface_l':
                self.model_file = model_dir + '/glint360_r100.onnx'
                if os.path.exists(model_dir + '/glint360_r100.onnx') == False:
                    gdown.download('https://drive.google.com/u/0/uc?id=1eHNQK7QbHFNEHqRUOlQNC0ju5E-ctjHX&export=download', model_dir + '/glint360_r100.onnx', quiet=False)
        else:
            self.model_file = model_file
        self.session = session
        self.taskname = 'recognition'
        find_sub = False
        find_mul = False
        model = onnx.load(self.model_file)
        graph = model.graph
        for nid, node in enumerate(graph.node[:8]):
            #print(nid, node.name)
            if node.name.startswith('Sub') or node.name.startswith('_minus'):
                find_sub = True
            if node.name.startswith('Mul') or node.name.startswith('_mul'):
                find_mul = True
        if find_sub and find_mul:
            #mxnet arcface model
            input_mean = 0.0
            input_std = 1.0
        else:
            input_mean = 127.5
            input_std = 127.5
        self.input_mean = input_mean
        self.input_std = input_std
        #print('input mean and std:', self.input_mean, self.input_std)
        if self.session is None:
            self.session = onnxruntime.InferenceSession(self.model_file, None)
        input_cfg = self.session.get_inputs()[0]
        input_shape = input_cfg.shape
        input_name = input_cfg.name
        self.input_size = tuple(input_shape[2:4][::-1])
        self.input_shape = input_shape
        outputs = self.session.get_outputs()
        output_names = []
        for out in outputs:
            output_names.append(out.name)
        self.input_name = input_name
        self.output_names = output_names
        assert len(self.output_names)==1
        self.output_shape = outputs[0].shape
    
    def get(self, img, kps):
        # print(onnxruntime.get_device())
        # cv2.imshow('d', img)
        aimg = face_align.norm_crop(img, landmark=kps)
        # cv2.imshow('aaasd', aimg)
        # cv2.waitKey(1)
        # cv2.imwrite('aaaa.jpg', aimg)
        embedding = self.get_feat(aimg).flatten()
        return embedding

    def get_feat(self, imgs):
        if not isinstance(imgs, list):
            imgs = [imgs]
        input_size = self.input_size
        
        blob = cv2.dnn.blobFromImages(imgs, 1.0 / self.input_std, input_size,
                                      (self.input_mean, self.input_mean, self.input_mean), swapRB=True)
        net_out = self.session.run(self.output_names, {self.input_name: blob})[0]
        return net_out


